#pragma once

#include <cassert>
#include <variant>

#include "./arena.hpp"
#include "tokenization.hpp"

struct NodeStmt;
struct NodeExpr;

//region Terms
struct NodeTermIntLit {
    Token int_lit;
};

struct NodeTermIdent {
    Token ident;
};

struct NodeTermParen {
    NodeExpr* expr;
};

//endregion

struct NodeTerm {
    std::variant<NodeTermIntLit*, NodeTermIdent*, NodeTermParen*> var;
};

//region BinExpr

//region Maths
struct NodeBinExprAdd {
    NodeExpr* lhs;
    NodeExpr* rhs;
};

struct NodeBinExprSub {
    NodeExpr* lhs;
    NodeExpr* rhs;
};

struct NodeBinExprMulti {
    NodeExpr* lhs;
    NodeExpr* rhs;
};

struct NodeBinExprDiv {
    NodeExpr* lhs;
    NodeExpr* rhs;
};

struct NodeBinExprPwr {
    NodeExpr* lhs;
    NodeExpr* rhs;
};
//endregion

//region Comparisons
struct NodeBinExprLT {
    NodeExpr* lhs;
    NodeExpr* rhs;
};
struct NodeBinExprLTEQ {
    NodeExpr* lhs;
    NodeExpr* rhs;
};

struct NodeBinExprGT {
    NodeExpr* lhs;
    NodeExpr* rhs;
};
struct NodeBinExprGTEQ {
    NodeExpr* lhs;
    NodeExpr* rhs;
};

struct NodeBinExprCmpr {
    NodeExpr* lhs;
    NodeExpr* rhs;
};
struct NodeBinExprCmprNEQ {
    NodeExpr* lhs;
    NodeExpr* rhs;
};
//endregion

//endregion

struct NodeBinExpr {
    std::variant<NodeBinExprCmprNEQ*, NodeBinExprAdd*, NodeBinExprCmpr* ,NodeBinExprGT* ,NodeBinExprGTEQ* ,NodeBinExprLT* ,NodeBinExprLTEQ*, NodeBinExprMulti*, NodeBinExprSub*, NodeBinExprDiv*, NodeBinExprPwr*> var;
};

struct NodeExpr {
    std::variant<NodeTerm*, NodeBinExpr*> var;
};

//region Statements

//region Assigning
struct NodeStmtAsgn {
    Token ident;
    NodeExpr* expr{};
};

struct NodeStmtInc {
    Token ident;
};
struct NodeStmtAddAsgn {
    Token ident;
    NodeExpr* expr;
};

struct NodeStmtDec {
    Token ident;
};
struct NodeStmtSubAsgn {
    Token ident;
    NodeExpr* expr;
};
//endregion

struct NodeStmtExit {
    NodeExpr* expr;
};

struct NodeStmtLet {
    Token ident;
    NodeExpr* expr{};
};

//region Loops
struct NodeStmtScope {
    std::vector<NodeStmt*> stmts;
};

struct NodeElse {
    NodeStmtScope* scope{};
};

struct NodeStmtIf {
    NodeExpr* expr{};
    NodeStmtScope* scope{};
    NodeElse* _else{};
};

struct NodeStmtWhile {
    NodeExpr* expr{};
    NodeStmtScope* scope{};
};
//endregion

struct NodeStmtCallFunc {
    Token ident;
};

struct NodeStmtDecFunc {
    Token ident;
    NodeStmtScope* scope{};
};

//endregion

struct NodeStmt {
    std::variant<NodeStmtCallFunc*, NodeStmtWhile*, NodeStmtDecFunc*, NodeStmtScope*, NodeStmtExit*, NodeStmtLet*, NodeStmtAsgn*, NodeStmtIf*, NodeStmtInc*, NodeStmtDec*, NodeStmtAddAsgn*, NodeStmtSubAsgn*> var;
};

struct NodeProg {
    std::vector<NodeStmt*> stmts;
};

class Parser {
public:
    inline explicit Parser(std::vector<Token> tokens)
        : m_tokens(std::move(tokens))
        , m_allocator(1024 * 1024 * 4) // 4 mb
    {
    }

    std::optional<NodeTerm*> parse_term()
    {

        if (auto int_lit = try_consume(TokenType::int_lit)) {
            auto term_int_lit = m_allocator.alloc<NodeTermIntLit>();
            term_int_lit->int_lit = int_lit.value();
            auto term = m_allocator.alloc<NodeTerm>();
            term->var = term_int_lit;
            return term;
        }
        else if (auto ident = try_consume(TokenType::ident)) {
            auto expr_ident = m_allocator.alloc<NodeTermIdent>();
            expr_ident->ident = ident.value();
            auto term = m_allocator.alloc<NodeTerm>();
            term->var = expr_ident;
            return term;
        }
        else if (auto open_paren = try_consume(TokenType::open_paren)) {
            auto expr = parse_expr();
            if (!expr.has_value()) {
                std::cerr << "Expected expression" << std::endl;
                exit(EXIT_FAILURE);
            }
            try_consume(TokenType::close_paren, "Expected `)`");
            auto term_paren = m_allocator.alloc<NodeTermParen>();
            term_paren->expr = expr.value();
            auto term = m_allocator.alloc<NodeTerm>();
            term->var = term_paren;
            return term;
        }
        else {
            return {};
        }
    }

    std::optional<NodeExpr*> parse_expr(int min_prec = 0)
    {
        std::optional<NodeTerm*> term_lhs = parse_term();
        if (!term_lhs.has_value()) {
            return {};
        }
        auto expr_lhs = m_allocator.alloc<NodeExpr>();
        expr_lhs->var = term_lhs.value();

        while (true) {
            std::optional<Token> curr_tok = peek();
            std::optional<int> prec;
            if (curr_tok.has_value()) {
                prec = bin_prec(curr_tok->type);
                if (!prec.has_value() || prec < min_prec) {
                    break;
                }
            }
            else {
                break;
            }
            Token op = consume();
            int next_min_prec = prec.value() + 1;
            auto expr_rhs = parse_expr(next_min_prec);
            if (!expr_rhs.has_value()) {
                std::cerr << "Unable to parse expression" << std::endl;
                exit(EXIT_FAILURE);
            }
            auto expr = m_allocator.alloc<NodeBinExpr>();
            auto expr_lhs2 = m_allocator.alloc<NodeExpr>();
            if (op.type == TokenType::plus) {
                auto add = m_allocator.alloc<NodeBinExprAdd>();
                expr_lhs2->var = expr_lhs->var;
                add->lhs = expr_lhs2;
                add->rhs = expr_rhs.value();
                expr->var = add;
            }
            else if (op.type == TokenType::plusplus) {

            }
            else if (op.type == TokenType::starstar) {
                auto pwr = m_allocator.alloc<NodeBinExprPwr>();
                expr_lhs2->var = expr_lhs->var;
                pwr->lhs = expr_lhs2;
                pwr->rhs = expr_rhs.value();
                expr->var = pwr;
            }
            else if (op.type == TokenType::cmpr) {
                auto cmpr = m_allocator.alloc<NodeBinExprCmpr>();
                expr_lhs2->var = expr_lhs->var;
                cmpr->lhs = expr_lhs2;
                cmpr->rhs = expr_rhs.value();
                expr->var = cmpr;
            }
            else if (op.type == TokenType::gt) {
                auto gt = m_allocator.alloc<NodeBinExprGT>();
                expr_lhs2->var = expr_lhs->var;
                gt->lhs = expr_lhs2;
                gt->rhs = expr_rhs.value();
                expr->var = gt;
            }
            else if (op.type == TokenType::gteq) {
                auto gteq = m_allocator.alloc<NodeBinExprGTEQ>();
                expr_lhs2->var = expr_lhs->var;
                gteq->lhs = expr_lhs2;
                gteq->rhs = expr_rhs.value();
                expr->var = gteq;
            }
            else if (op.type == TokenType::lt) {
                auto lt = m_allocator.alloc<NodeBinExprLT>();
                expr_lhs2->var = expr_lhs->var;
                lt->lhs = expr_lhs2;
                lt->rhs = expr_rhs.value();
                expr->var = lt;
            }
            else if (op.type == TokenType::lteq) {
                auto lteq = m_allocator.alloc<NodeBinExprLTEQ>();
                expr_lhs2->var = expr_lhs->var;
                lteq->lhs = expr_lhs2;
                lteq->rhs = expr_rhs.value();
                expr->var = lteq;
            }
            else if (op.type == TokenType::star) {
                auto multi = m_allocator.alloc<NodeBinExprMulti>();
                expr_lhs2->var = expr_lhs->var;
                multi->lhs = expr_lhs2;
                multi->rhs = expr_rhs.value();
                expr->var = multi;
            }
            else if (op.type == TokenType::sub) {
                auto sub = m_allocator.alloc<NodeBinExprSub>();
                expr_lhs2->var = expr_lhs->var;
                sub->lhs = expr_lhs2;
                sub->rhs = expr_rhs.value();
                expr->var = sub;
            }
            else if (op.type == TokenType::div) {
                auto div = m_allocator.alloc<NodeBinExprDiv>();
                expr_lhs2->var = expr_lhs->var;
                div->lhs = expr_lhs2;
                div->rhs = expr_rhs.value();
                expr->var = div;
            }
            else if (op.type == TokenType::noteq) {
                auto cmpr = m_allocator.alloc<NodeBinExprCmprNEQ>();
                expr_lhs2->var = expr_lhs->var;
                cmpr->lhs = expr_lhs2;
                cmpr->rhs = expr_rhs.value();
                expr->var = cmpr;
            }
            else {
                assert(false); // Unreachable;
            }
            expr_lhs->var = expr;
        }
        return expr_lhs;
    }

    std::optional<NodeStmt*> parse_stmt()
    {
        // exit
        if (peek().value().type == TokenType::exit && peek(1).has_value() && peek(1).value().type == TokenType::open_paren) {
            consume();
            consume();
            auto stmt_exit = m_allocator.alloc<NodeStmtExit>();
            if (auto node_expr = parse_expr()) {
                stmt_exit->expr = node_expr.value();
            }
            else {
                std::cerr << "Invalid expression" << std::endl;
                exit(EXIT_FAILURE);
            }
            try_consume(TokenType::close_paren, "Expected `)`");
            try_consume(TokenType::semi, "Expected `;`");
            auto stmt = m_allocator.alloc<NodeStmt>();
            stmt->var = stmt_exit;
            return stmt;
        }
        // if
        else if (peek().value().type == TokenType::_if) {
            return parse_if();
        }
        // while
        else if (peek().value().type == TokenType::_while) {
            return parse_while();
        }
        // (expr)
        else if (peek().value().type == TokenType::ident && peek(1).has_value() && peek(1).value().type == TokenType::eq) {
            auto identToken = consume(); // Save the identifier token
            consume(); // Consume the "=" token
            auto stmt_asgn = m_allocator.alloc<NodeStmtAsgn>();
            stmt_asgn->ident = identToken; // Set the identifier in the stmt_asgn
            if (auto node_expr = parse_expr()) {
                stmt_asgn->expr = node_expr.value();
            }
            else {
                std::cerr << "Invalid expression" << std::endl;
                exit(EXIT_FAILURE);
            }
            try_consume(TokenType::semi, "Expected `;`");
            auto stmt = m_allocator.alloc<NodeStmt>();
            stmt->var = stmt_asgn;
            return stmt;
        }
        // ++
        else if (peek().value().type == TokenType::ident && peek(1).has_value() && peek(1).value().type == TokenType::plusplus) {
            auto identToken = consume(); // Save the identifier token
            consume(); // Consume the "++" token
            auto stmt_inc = m_allocator.alloc<NodeStmtInc>();
            stmt_inc->ident = identToken; // Set the identifier in the stmt_asgn
            try_consume(TokenType::semi, "Expected `;`");
            auto stmt = m_allocator.alloc<NodeStmt>();
            stmt->var = stmt_inc;
            return stmt;
        }
        // +=
        else if (peek().value().type == TokenType::ident && peek(1).has_value() && peek(1).value().type == TokenType::pluseq) {
            auto identToken = consume(); // Save the identifier token
            //std::cerr << "here" << std::endl;
            consume(); // Consume the "+=" token
            auto addasgn = m_allocator.alloc<NodeStmtAddAsgn>();
            addasgn->ident = identToken; // Set the identifier in the stmt_asgn

            if (auto node_expr = parse_expr()) {
                addasgn->expr = node_expr.value();
            }
            else {
                std::cerr << "Invalid expression" << std::endl;
                exit(EXIT_FAILURE);
            }
            try_consume(TokenType::semi, "Expected `;`");
            auto stmt = m_allocator.alloc<NodeStmt>();
            stmt->var = addasgn;
            return stmt;
        }
        // --
        else if (peek().value().type == TokenType::ident && peek(1).has_value() && peek(1).value().type == TokenType::minusminus) {
            auto identToken = consume(); // Save the identifier token
            consume(); // Consume the "--" token
            auto stmt_dec = m_allocator.alloc<NodeStmtDec>();
            stmt_dec->ident = identToken; // Set the identifier in the stmt_asgn
            try_consume(TokenType::semi, "Expected `;`");
            auto stmt = m_allocator.alloc<NodeStmt>();
            stmt->var = stmt_dec;
            return stmt;
        }
        // -=
        else if (peek().value().type == TokenType::ident && peek(1).has_value() && peek(1).value().type == TokenType::minuseq) {
            auto identToken = consume(); // Save the identifier token
            consume(); // Consume the "-=" token
            auto subasgn = m_allocator.alloc<NodeStmtSubAsgn>();
            subasgn->ident = identToken; // Set the identifier in the stmt_asgn
            try_consume(TokenType::semi, "Expected `;`");
            auto stmt = m_allocator.alloc<NodeStmt>();
            stmt->var = subasgn;
            return stmt;
        }
        // let =
        else if (peek().has_value() && peek().value().type == TokenType::let && peek(1).has_value() && peek(1).value().type == TokenType::ident && peek(2).has_value() && peek(2).value().type == TokenType::eq) {
            consume();
            auto stmt_let = m_allocator.alloc<NodeStmtLet>();
            stmt_let->ident = consume();
            consume();
            if (auto expr = parse_expr()) {
                stmt_let->expr = expr.value();
            }
            else {
                std::cerr << "Invalid expression" << std::endl;
                exit(EXIT_FAILURE);
            }
            try_consume(TokenType::semi, "Expected `;`");
            auto stmt = m_allocator.alloc<NodeStmt>();
            stmt->var = stmt_let;
            return stmt;
        }
        // scope
        else if (peek().has_value() && peek().value().type == TokenType::open_curly) {
            if (auto scope = parse_scope()) {
                auto stmt = m_allocator.alloc<NodeStmt>();
                stmt->var = scope.value();
                return stmt;
            }
            else {
                std::cerr << "Invalid scope" << std::endl;
                exit(EXIT_FAILURE);
            }
        }
        // declare function
        else if (peek().has_value() && peek().value().type == TokenType::func) {

            if (peek(4).value().type == TokenType::open_curly) {
                return parse_func();
            }
        }
        else if (peek().value().type == TokenType::ident && peek(1).has_value() && peek(1).value().type == TokenType::open_paren) {
            return parse_callfunc();
        }
        else {
            return {};
        }
    }

    std::optional<NodeStmtScope*> parse_scope()
    {
        if (!try_consume(TokenType::open_curly).has_value()) {
            return {};
        }
        auto scope = m_allocator.alloc<NodeStmtScope>();
        while (auto stmt = parse_stmt()) {
            scope->stmts.push_back(stmt.value());
        }
        try_consume(TokenType::close_curly, "Expected `}`");
        return scope;
    }

    std::optional<NodeStmt*> parse_func() {
        consume(); // Consume the "func" token

        // Consume function name
        auto identToken = try_consume(TokenType::ident);
        if (!identToken.has_value()) {
            std::cerr << "Expected function name identifier" << std::endl;
            exit(EXIT_FAILURE);
        }

        try_consume(TokenType::open_paren, "Expected `(`");
        //parameters go here
        try_consume(TokenType::close_paren, "Expected `)`");

        NodeStmtScope* funcScope;
        if (auto scope = parse_scope()) {
            funcScope = scope.value();
        } else {

        }

        // Create function node and populate fields
        auto funcNode = m_allocator.alloc<NodeStmtDecFunc>();
        funcNode->ident = identToken.value();
        funcNode->scope = funcScope;

        // Wrap function node in statement node and return
        auto stmt = m_allocator.alloc<NodeStmt>();
        stmt->var = funcNode;  // Assign the parsed NodeStmtFunc* to the var member of NodeStmt
        return stmt;  // Return the NodeStmt*
    }

    std::optional<NodeStmt*> parse_callfunc() {
        // Consume function name
        auto identToken = try_consume(TokenType::ident);
        if (!identToken.has_value()) {
            std::cerr << "Expected function name identifier" << std::endl;
            exit(EXIT_FAILURE);
        }

        try_consume(TokenType::open_paren, "Expected `(`");
        //parameters go here
        try_consume(TokenType::close_paren, "Expected `)`");
        try_consume(TokenType::semi, "Expected `;`");

        // Create function node and populate fields
        auto funcNode = m_allocator.alloc<NodeStmtCallFunc>();
        funcNode->ident = identToken.value();

        // Wrap function node in statement node and return
        auto stmt = m_allocator.alloc<NodeStmt>();

        stmt->var = funcNode;  // Assign the parsed NodeStmtFunc* to the var member of NodeStmt
        return stmt;  // Return the NodeStmt*
    }

    std::optional<NodeProg> parse_prog()
    {
        NodeProg prog;
        while (peek().has_value()) {
            if (auto stmt = parse_stmt()) {
                prog.stmts.push_back(stmt.value());
            }
            else {
                std::cerr << "Invalid statement" << std::endl;
                exit(EXIT_FAILURE);
            }
        }
        return prog;
    }

    std::optional<NodeStmt*> parse_if(){
        consume(); // Consume the "if" token
        try_consume(TokenType::open_paren, "Expected `(` before condition");
        auto stmt_if = m_allocator.alloc<NodeStmtIf>();
        if (auto node_expr = parse_expr()) {
            stmt_if->expr = node_expr.value();
        }
        else {
            std::cerr << "Invalid expression" << std::endl;
            exit(EXIT_FAILURE);
        }

        try_consume(TokenType::close_paren, "Expected `)` after condition");


        if (auto scope = parse_scope()) {
            stmt_if->scope = scope.value();
        } else {
            std::cerr << "Invalid Scope" << std::endl;
            exit(EXIT_FAILURE);
        }

        if (peek().has_value() && peek().value().type == TokenType::_else) {
            auto else_node = m_allocator.alloc<NodeElse>();
            consume();
            if (auto scope = parse_scope()) {
                else_node->scope = scope.value();
            } else {
                std::cerr << "Invalid Scope" << std::endl;
                exit(EXIT_FAILURE);
            }
            //try_consume(TokenType::open_curly, "Expected `{` after else");
            // auto else_node = m_allocator.alloc<NodeElse>();
            //else_node->stmts = parse_nestedstmts();
            //try_consume(TokenType::close_curly, "Expected `}` after else block");
            stmt_if->_else = else_node;
        }

        auto stmt = m_allocator.alloc<NodeStmt>();
        stmt->var = stmt_if;
        return stmt;
    }

    std::optional<NodeStmt*> parse_while(){
        consume(); // Consume the "while" token
        try_consume(TokenType::open_paren, "Expected `(` before condition");
        auto stmt_while = m_allocator.alloc<NodeStmtWhile>();
        if (auto node_expr = parse_expr()) {
            stmt_while->expr = node_expr.value();
        }
        else {
            std::cerr << "Invalid expression" << std::endl;
            exit(EXIT_FAILURE);
        }

        try_consume(TokenType::close_paren, "Expected `)` after condition");


        if (auto scope = parse_scope()) {
            stmt_while->scope = scope.value();
        } else {
            std::cerr << "Invalid Scope" << std::endl;
            exit(EXIT_FAILURE);
        }

        auto stmt = m_allocator.alloc<NodeStmt>();
        stmt->var = stmt_while;
        return stmt;
    }

private:
    [[nodiscard]] inline std::optional<Token> peek(int offset = 0) const
    {
        if (m_index + offset >= m_tokens.size()) {
            return {};
        }
        else {
            return m_tokens.at(m_index + offset);
        }
    }

    inline Token consume()
    {
        Token currToken = m_tokens.at(m_index++);

        // Print out the consumed token's type
        //std::cout << "Consumed: " << TokenTypeToString(currToken.type) << std::endl;

        return currToken;

    }

    const static char* TokenTypeToString(TokenType type) {
        switch (type) {
        case TokenType::_if: return "If\n";
        case TokenType::_while: return "While\n";
        case TokenType::int_lit: return "Number\n";
        case TokenType::semi: return "';'\n";
        case TokenType::open_paren: return "'('\n";
        case TokenType::close_paren: return "')'\n";
        case TokenType::open_curly: return "'{'\n";
        case TokenType::close_curly: return "'}'\n";
        case TokenType::ident: return "'Variable'\n";
        case TokenType::let: return "'Let'\n";
        case TokenType::_else: return "'Else'\n";
        case TokenType::eq: return "'='\n";
        case TokenType::plus: return "'+'\n";
        case TokenType::star: return "'*'\n";
        case TokenType::sub: return "'-'\n";
        case TokenType::div: return "'/'\n";
        case TokenType::starstar: return "'**'\n";
        case TokenType::cmpr: return "'=='\n";
        case TokenType::pluseq: return "'+='\n";
        case TokenType::minuseq: return "'-='\n";
        case TokenType::gt: return "'<'\n";
        case TokenType::lt: return "'>'\n";
        case TokenType::gteq: return "'<='\n";
        case TokenType::lteq: return "'>='\n";
        case TokenType::func: return "Function\n";
        default: return "Unknown TokenType";
        }
    }

    inline Token try_consume(TokenType type, const std::string& err_msg)
    {
        if (peek().has_value() && peek().value().type == type) {
            return consume();
        }
        else {
            std::cerr << err_msg << std::endl;
            exit(EXIT_FAILURE);
        }
    }

    inline std::optional<Token> try_consume(TokenType type)
    {
        if (peek().has_value() && peek().value().type == type) {
            return consume();
        }
        else {
            return {};
        }
    }

    const std::vector<Token> m_tokens;
    size_t m_index = 0;
    ArenaAllocator m_allocator;
};